/* Plot3DView.m   Copyright 1992 Steve Ludtke */
/* This view allows the display of 3d functions with real time rotation */

#import "Plot3DView.h"
#import "PlotShape.h"
#import "PControl.h"
#import <appkit/appkit.h>
#import <3Dkit/3Dkit.h>
#import <dpsclient/event.h>
#import <dpsclient/psops.h>
#import <stdio.h>
#import <math.h>
#include <libc.h>
#import "Expression.h"
#import "DensView.h"

void PSsethel();	/* makes Helvetica current font */

#define LRAD 10.0

extern id NXApp;

@implementation Plot3DView
-initFrame:(NXRect *)myrect
{
int i;
  RtPoint fromP = {0,0,6.0}, toP = {0,0,0};
  RtPoint lFromP = {5.0,10.0,5.0},lToP = {0,0,0};
  id aShader;

[super initFrame:(NXRect *)myrect];

[self setBackgroundColor:NX_COLORWHITE];  
[self setDrawBackgroundColor:YES];

  [self setEyeAt:fromP toward:toP roll:0.0];
  
  aShader=[[N3DShader alloc] init];
  [(N3DShader *)aShader setShader:"matte"];

  shape=[[PlotShape alloc] init];
  [(N3DShape *) shape  setShader:aShader];
  [shape scale:-1.0 :1.0 :1.0];
  [[self setWorldShape:shape] free];

  ambLight=[[N3DLight alloc] init];
  [ambLight makeAmbientWithIntensity:0.1];
  [self addLight:ambLight];


ltheta=.785;
lchi=-.785;
aLight=[[N3DLight alloc] init];
lFromP[0]= LRAD*sin(lchi)*cos(ltheta);
lFromP[2]= LRAD*cos(lchi)*cos(ltheta);
lFromP[1]= LRAD*sin(ltheta);
[aLight makeDistantFrom:lFromP to:lToP intensity:0.9];
[self addLight:aLight];

  Rmode=3;

/* initialize preferences */
for (i=0; i<MAXSETS; i++) {
	pref[i].ndata=0; 
	pref[i].sym= -1;
	pref[i].color[0]=pref[i].color[1]=pref[i].color[2]=1.0;
	pref[i].lcolor[0]=pref[i].lcolor[1]=pref[i].lcolor[2]=0.0;
	pref[i].fileData=NULL;
	pref[i].expr=[[Expression alloc] init];
	[pref[i].expr parse:"0"];
	pref[i].nx=pref[i].ny=10;
}
/* initial function to display */
/*[pref[0].expr parse:"cos(sqrt(x^2+y^2))/(sqrt(x^2+(y/3)^2)+1)"];*/
pref[0].sym=5;
pref[0].nx=pref[0].ny=20;

/* start spinning, all angles in radians */
chi=0;
theta=.733;
/*dchi=0.025;*/
dchi=0.0;
initflag=1;
Rflags=1;

[[[NXApp printInfo] setVertCentered:YES] setOrientation:NX_PORTRAIT andAdjust:YES];
[[NXApp printInfo] setMarginLeft:0.0 right:0.0 top:0.0 bottom:0.0];
[[NXApp printInfo] setHorizPagination:NX_FITPAGINATION];
[[NXApp printInfo] setVertPagination:NX_FITPAGINATION];

timer=(DPSTimedEntry)DPSAddTimedEntry(TIMESTEP,itstime,self,NX_RUNMODALTHRESHOLD);

return self;
}

/* fix the scaling after being resized */
-superviewSizeChanged:(const NXSize *)oldsize
{
[super superviewSizeChanged:oldsize];
[self zoom:self];
return self;
}

-free
{
DPSRemoveTimedEntry(timer);
[super free];
return self;
}

-drawPS:(NXRect *)myrect :(int)rectCount
{
char s[80];

if (bounds.size.width<150) return self;
/* display alt and az in upper right corner */
PSsetgray(0.0);
PSsethel();
sprintf(s,"Alt:%6.2f  Az:%6.2f",theta*57.295787,chi*57.295787);
PSmoveto(bounds.size.width-100.0,bounds.size.height-15.0);
PSshow(s);
PSstroke();

return self;
}	

/* recalculates and redisplays data */
-zoom:sender
{
float x0,x1,xs,y0,y1,ys,xf,yf,zf;
float z0,z1,x,y,zz;
int i,j=0,k;
id tmp;

/* calculate minima, maxima and step sizes */
x0=minX;
x1=maxX;
y0=minY;
y1=maxY;

xf=(x1-x0)/2.0;
yf=(y1-y0)/2.0;

/* make sure there's enough space in the data array */
for (i=0; i<NFN; i++) {
	if (pref[i].sym==-1 || pref[i].fileData!=NULL) continue;
	if (pref[i].ndata!=(pref[i].nx*pref[i].ny)) {
		if (pref[i].ndata!=0) free(pref[i].data);
		pref[i].data=malloc(sizeof(Point)*(pref[i].nx*pref[i].ny+2));
		pref[i].ndata=pref[i].nx*pref[i].ny;
	}
}
z0=MAXFLOAT;
z1= -MAXFLOAT;

/* calculate (function mode) or clip (file mode) the data */
/* also finds min/max Z */
for (i=0; i<NFN; i++) {
	if (pref[i].sym==-1) continue;
	tmp=pref[i].expr;
	if (pref[i].fileData!=NULL) {
		k=0;
		for (j=0; j<pref[i].nfdata; j++) {
			if (pref[i].fileData[j].x>x1||pref[i].fileData[j].x<x0||
			    pref[i].fileData[j].y>y1||pref[i].fileData[j].y<y0) continue;
			pref[i].data[k].x=(pref[i].fileData[j].x-x0)/xf-1.0;
			[tmp setVar:"x" value:pref[i].data[k].x];
			pref[i].data[k].y=(pref[i].fileData[j].y-y0)/yf-1.0;
			[tmp setVar:"y" value:pref[i].data[k].y];
			[tmp setVar:"z" value:pref[i].fileData[j].z];
			pref[i].data[k].z=[tmp resultValue];
			if (pref[i].data[k].z>z1) z1=pref[i].data[k].z;
			if (pref[i].data[k].z<z0) z0=pref[i].data[k].z;
			k++;
		}
		pref[i].ndata=k;

		if (pref[i].data[0].y!=pref[i].data[1].y) {
			pref[i].nx=pref[i].ny=(int)floor(sqrt((float)k));
		}
		else {
			for (j=1; j<k; j++) if (pref[i].data[j-1].y!=pref[i].data[j].y) break;
			pref[i].nx=j;
			pref[i].ny=k/j;
		}
		continue;
	}

	xs=(maxX-x0)/(float)(pref[i].nx-1);
	ys=(maxY-y0)/(float)(pref[i].ny-1);	
	x1=maxX+xs/2.0;
	y1=maxY+ys/2.0;
	j=0;
	for (y=y0; y<y1; y+=ys) {
		for (x=x0; x<x1; x+=xs) {
			pref[i].data[j].x=(x-x0)/xf-1.0;
			pref[i].data[j].y=(y-y0)/yf-1.0;
			[tmp setVar:"x" value:x];
			[tmp setVar:"y" value:y];
			zz=pref[i].data[j].z=[tmp resultValue];
			if (zz!=zz || zz<-MAXFLOAT || zz>MAXFLOAT) zz=pref[i].data[j].z=0;
			if (zz>z1) z1=zz;
			if (zz<z0) z0=zz;
			j++;
		}
	}
}

if (z1<=z0) { z1=z0+.001; z0-=.001; }
/* allow the controller to override min/max Z */
[controller minmaxZ:&z0 :&z1];

zf=(z1-z0)/2.0;
/* scale and clip Z */
for (i=0; i<NFN; i++) {
	if (pref[i].sym==-1) continue;
	for (k=0; k<pref[i].ndata; k++) {
		if (pref[i].data[k].z>z1) pref[i].data[k].z=z1;
		if (pref[i].data[k].z<z0) pref[i].data[k].z=z0;
		pref[i].data[k].z=(pref[i].data[k].z-z0)/zf-1.0;
	}
}

/* display 3d plot */
[shape setData:pref :Rmode :Rflags];
[shape setAng:theta :chi];
[self display];
/* display density plot */
[controller updDen];
return self;
}

/* obvious */
-setAng:(float)az :(float)alt
{
theta=alt;
chi=az;
return self;
}

/* allows spinning and zooming with the mouse */
-mouseDown:(NXEvent *)oevent 
{
int oldMask,loop=1;
float ix=0,iy=0,ix1=0,iy1=0,ix2=0,iy2=0,ix3=0,iy3=0;
NXEvent *event,evs;
long tm,tm2;

evs=*oevent;
oevent=&evs;
[self convertPoint:&oevent->location fromView:nil];
oevent->location.x=oevent->location.x/bounds.size.width*2.0-1.0;
oevent->location.y=oevent->location.y/bounds.size.height*2.0-1.0;
ix2=ix=oevent->location.x;
iy2=iy=oevent->location.y;
tm2=tm=oevent->time;

oldMask = [window addToEventMask:NX_LMOUSEDRAGGEDMASK];

while (loop) {
    event = [NXApp getNextEvent:(NX_LMOUSEUPMASK | NX_LMOUSEDRAGGEDMASK)];
    [self convertPoint:&event->location fromView:nil];
	event->location.x=event->location.x/bounds.size.width*2.0-1.0;
	event->location.y=event->location.y/bounds.size.height*2.0-1.0;

        switch (event->type) {
        case NX_LMOUSEUP:
                loop = 0;
	        	dchi=-(event->location.x-ix2)/(float)(event->time-tm2)*15.0;
				if (fabs(dchi)<.004) dchi=0.0;
				[shape setAng:theta :chi];
                [self display];
            break;
        case NX_LMOUSEDRAGGED:
	        theta-=(event->location.y-iy)*3.0;
	        chi-=(event->location.x-ix)*4.0;
	        if (theta>M_PI/2.0) theta=M_PI/2.0;
	        if (theta<-M_PI/2.0) theta=-M_PI/2.0;
	        while (chi>2.0*M_PI) chi-=2.0*M_PI;
	        while (chi<0.0) chi+=2.0*M_PI;
	        ix2=ix;
	        iy2=iy;
	        tm2=tm;
	        ix=event->location.x;
	        iy=event->location.y;
	        tm=event->time;
			[shape setAng:theta :chi];
	        [self display];
            break;
        default:
            break;
        }
}
[window setEventMask:oldMask];
return self;
}

/* function called by timer */
void itstime(DPSTimedEntry entry,double now,id call)
{
[call step];
return;
}

/* do one time step */
-step
{
if (initflag) { 
	/* first time, do initialization stuff */
	if (controller==nil) return self;
	[controller startup:pref];
	[self setSurfaceTypeForAll:N3D_SmoothSolids chooseHider:YES];
	[self zoom:self];
	initflag=0; 
}
if (dchi==0.0) return self;
chi+=dchi;
while (chi>2.0*M_PI) chi-=2.0*M_PI;
while (chi<0.0) chi+=2.0*M_PI;
[shape setAng:theta :chi];
[self display];
return self;
}

/* new min/max values */
-zoomTo:(float)minx :(float)miny :(float)maxx :(float)maxy
{
minX=minx;
minY=miny;
maxX=maxx;
maxY=maxy;
return self;
}

-(int)acceptsFirstMouse { return (YES); }

-setcontroller:con 
{
controller=con;
return self;
}

/* allow user to pause spinning */
-togFreeze:sender
{
static float Tdchi;

if ([sender intValue]) {
	Tdchi=dchi;
	dchi=0.0;
	return self;
}
dchi=Tdchi;
return self;
}

- renderSelf:(RtToken)context
{
/*RiErrorIgnore();*/
return self;
}

- dumpRib:sender
{
  static id savePanel=nil;
  NXStream *ts;
  char buf[MAXPATHLEN+1];

  if (!savePanel) {
    savePanel=[SavePanel new];
    [savePanel setRequiredFileType:"rib"];
  }

  if([savePanel runModal]){
    ts=NXOpenMemory(NULL, 0, NX_WRITEONLY);
    strcpy(buf, [savePanel filename]);
    strrchr(buf,'.')[0]='\0';
    NXPrintf(ts, "Display \"%s.tiff\" \"file\" \"rgba\"\n", buf);
    [self copyRIBCode:ts];
    NXSaveToFile(ts, [savePanel filename]);
    NXCloseMemory(ts,NX_FREEBUFFER);
  }
  return self;
}

- setAmbLight:sender
{
[ambLight setIntensity:[sender floatValue]];
[self display];
return self;
}

- setLight:sender
{
[aLight setIntensity:[sender floatValue]];
[self display];
return self;
}

-setLightX:sender
{
RtPoint from;
lchi= -[sender floatValue];
from[0]= LRAD*sin(lchi)*cos(ltheta);
from[2]= LRAD*cos(lchi)*cos(ltheta);
from[1]= LRAD*sin(ltheta);
[aLight setFrom:from];
[self display];
return self;
}

-setLightY:sender
{
RtPoint from;
ltheta=[sender floatValue];
from[0]= LRAD*sin(lchi)*cos(ltheta);
from[2]= LRAD*cos(lchi)*cos(ltheta);
from[1]= LRAD*sin(ltheta);
[aLight setFrom:from];
[self display];
return self;
}

- setMode:sender
{
int i;

Rmode=i=[sender selectedRow];
switch(i) {
case 0:
 [self setSurfaceTypeForAll:N3D_PointCloud chooseHider:YES];
 [window setDepthLimit:NX_TwoBitGrayDepth];
break;
case 1:
 [self setSurfaceTypeForAll:N3D_WireFrame chooseHider:YES];
 [window setDepthLimit:NX_TwoBitGrayDepth];
break;
case 2:
 [self setSurfaceTypeForAll:N3D_FacetedSolids chooseHider:YES];
break;
case 3:
 [self setSurfaceTypeForAll:N3D_SmoothSolids chooseHider:YES];
break;
}
[shape setData:pref :Rmode :Rflags];
[self display];
return self;
}

-printPSCode:sender
{
[shape setData:pref :Rmode+4 :Rflags];
[super printPSCode:sender];
[shape setData:pref :Rmode :Rflags];
return self;
}

-setFlags:sender
{
Rflags=0;
if ([[flagSel findCellWithTag:0] intValue]) Rflags+=1;
if ([[flagSel findCellWithTag:1] intValue]) Rflags+=2;
[shape setData:pref :Rmode :Rflags];
[self display];
return self;
}
@end
